# encoding: utf-8
require "yaml"
require "kramdown"
require "fileutils"
require "nkf"
require "kpdf"
require "wrdap"

    ## function definition
def yencode(hs, enc=nil)
  enc = __ENCODING__  unless enc
  hs.each do |key, val|
    if val.class == String
      hs[key] = val.encode(enc, :invalid => :replace)
    elsif val.class == Array
      for i in 0...val.size
        if val[i].class == String
          val[i] = val[i].encode(enc, :invalid => :replace)
        end
      end
      hs[key] = val
    elsif val.class == Hash
      hs[key] = yencode(val)
    end
  end
  return hs
end

def ysearch(hs, skey)
  res = nil
  if hs.class != Hash
    return res
  end
  if hs[skey] != nil
    res = hs[skey]
  else
    hs.each do |key, val|
      if val.class == Hash
        res = ysearch(val, skey)
      end
      break  if res != nil
    end
  end
  return res
end

def ydoctype(hs)
  doc_type = %w(html_document md_document pdf_document) +
      %w(tex_document kpdf_document word_document)
  res = "html_document"
  if hs.class != Hash or !hs['output']
    return res
  end
  if doc_type.include?(hs['output'])
    res = hs['output']
  elsif hs['output'].class == Hash
    hs['output'].each do |key, val|
      if doc_type.include?(key)
        res = key
        break
      end
    end
  end
  return res
end

def output_move(bname, flist)
  flist2 = {}
  Dir.glob("./*") do |fname|
    next  if test(?d, fname)
    fkey = File.basename(fname)
    fobj = File::stat(fname)
    flist2[fkey] = fobj.mtime.to_s
  end
  dname = bname + "_output"
  Dir.mkdir(dname)  unless test(?e, dname)
  if test(?d, dname)
    Dir.glob("./#{dname}/*") {|fname|  File.unlink(fname)}
    flist2.each do |fkey, fmod|
      if !flist[fkey] or fmod > flist[fkey]
        FileUtils.mv(fkey, "./#{dname}")  rescue dummy=nil
      end
    end
  end
end

def file_read(fname)
  str = File.read(fname).to_s
  return NKF.nkf("-m0w", str)
end

def conv_tex(str)
  str = str.gsub(/[\^\\<~|#$%&_{}>]/) {|ptn|
    case ptn
    when /[#$%&_{}>]/
      ptn = "\\" + ptn
    when /[\^<~]/
      ptn = "\\verb|#{ptn}|"
    when /\\/
      ptn = "\\verb|\\|"
    when /\|/
      ptn = "\\verb+|+"
    end
  }
  return str
end

    ## prepare
mathjax_scr, tpl_htm, tpl_tex =
    DATA.read.gsub(/\r\n/, "\n").split(/^####----####\n/)
$headpart = ''
tpl_file = "$$KramdownTemplate.txt"
flist = {}  # カレントディレクトリのファイルリストを取得
Dir.glob("./*") do |fname|
  next  if test(?d, fname)
  fkey = File.basename(fname)
  fobj = File::stat(fname)
  flist[fkey] = fobj.mtime.to_s
end

    ## main
unless test(?e, ARGV[0])
  STDERR.puts "'#{ARGV[0]}' がみつかりません."
  exit
end

start_time = Time.now.to_f
rmd_file = File.basename(ARGV[0])
bname = File.basename(rmd_file, ".*")
md_file = bname + ".md"
tempb = "$$MrmdTemp"
log_file = "log.txt"

    ## markdownファイルの生成・読み込み・調整
wld = (ENV['RPT_WITH'] + "\\*.r").gsub(/\\/, "/")
src_str = ''
Dir.glob(wld) do |fname|
  src_str += "source(\"#{fname}\")\n"
end
r_scr = <<EOS
library(knitr)
#{src_str}\
infile.rmd <- "#{rmd_file}"
outfile.md <- "#{md_file}"
knit(input=infile.rmd, output=outfile.md)
EOS

if md_file != rmd_file
  r_file = "#{tempb}.R"
  File.open(r_file, "w") {|ff|  ff.print r_scr}
  File.delete(md_file)  if test(?e, md_file)
  File.delete(log_file)  if test(?e, log_file)
  system("rcmd BATCH --no-save -q --internet2 #{r_file} #{log_file}")
  Dir.glob("#{tempb}.*") {|fname|  File.delete(fname)}
end
unless test(?e, md_file)
  STDERR.puts "'#{md_file}' がみつかりません."
  exit
end
str_org = file_read(md_file)
str = str_org.dup

    ## yaml部分の読み込みとその調整
yopt = nil
if str =~ /\A---\n(.+?)\n---\n(.+)\Z/m
  ypart = $1 + "\n"
  str = $2
  yopt = yencode(YAML.load(ypart))
end
doc_type = ydoctype(yopt)

kopt = {}
kramdown_o = ysearch(yopt, 'kramdown_args')
if kramdown_o.class == Hash
  kramdown_o.each do |key, val|
    kopt[key] = val
  end
end
if kopt['template'] == false
  kopt.delete('template')
elsif kopt['template'].class != String
  kopt['template'] = tpl_file
end

popt = []
pandoc_o = ysearch(yopt, 'pandoc_args')
if pandoc_o.class == Hash
  pandoc_o.each do |key, val|
    popt << ((val == nil) ? key : {key=>val})
  end
elsif pandoc_o.class == Array
  pandoc_o.each do |elm|
    popt << elm
  end
end
if !popt.include?('toc') and ysearch(yopt, 'toc')
  popt << 'toc'
end

init_str = ''
if yopt
  case doc_type
  when "html_document", "md_document" then
    if yopt['title']
      init_str += "# #{yopt['title']}\n{: .title .no_toc #title}\n\n"
    end
    if yopt['author']
      init_str += "#### *#{yopt['author']}*\n{: .author .no_toc}\n\n"
    end
    if yopt['date']
      init_str += "#### *#{yopt['date']}*\n{: .date .no_toc}\n\n"
    end
    if init_str != ''
      init_str = "<!-- yaml block start -->\n\n" + init_str +
        "<!-- yaml block end -->\n\n--------\n\n"
    end
    toc_o = ysearch(yopt, 'toc')
    if toc_o
      ctitle = (toc_o.class == String) ? toc_o : "＜目　次＞"
      init_str += "## #{ctitle}\n{: .no_toc #toc}\n\n" +
        "* contents list\n{:toc}\n\n--------\n\n"
    end
    str = init_str + str
  when "kpdf_document" then
    if yopt['title']
      init_str += "# #{yopt['title']}\n\n"
    end
    if yopt['author']
      init_str += "#### *#{yopt['author']}*\n\n"
    end
    if yopt['date']
      init_str += "#### *#{yopt['date']}*\n\n"
    end
    init_str += "--------\n\n"  if init_str != ''
    str = init_str + str
  when "tex_document", "pdf_document" then
    toc_o = ysearch(yopt, 'toc')
    if toc_o
      init_str += "* contents list\n{:toc}\n\n--------\n\n"
    end
    str = init_str + str
  when "word_document" then
    if yopt['title']
      init_str += "% #{yopt['title']}\n"
    end
    if yopt['author']
      init_str += "% #{yopt['author']}\n"
    end
    if yopt['date']
      init_str += "% #{yopt['date']}\n"
    end
    if init_str != ''
      init_str += "\n"
    end
    str = init_str + str
  end
end

md_str = str.gsub(/^```+/, "~~~")
if md_str != str_org
  outfile = md_file
  outfile = outfile.gsub(/\./, "_") + ".md"  if md_file == rmd_file
  File.open(outfile, "w") {|ff|  ff.print md_str}
end

case doc_type
when "html_document" then
  mathjax_flag = ysearch(yopt, 'mathjax')
  unless mathjax_flag
    if md_str =~ /\$\$.+?\$\$|\\\(.+?\\\)|\\\[.+?\\\]/m
      mathjax_flag = true
    end
    if mathjax_flag
      kopt['math_engine'] = 'mathjax'
      $headpart += mathjax_scr
    end
  end
  if (whdpt = ysearch(yopt, 'headpart'))
    hdpt = ''
    if whdpt.class == String
      hdpt = test(?f, whdpt) ? file_read(whdpt) : whdpt
    elsif whdpt.class == Array
      hdpt = whdpt.join("\n") + "\n"
    end
    $headpart += hdpt
  end
  if kopt['template'] == tpl_file
    File.open(tpl_file, "w") {|ff|  ff.print tpl_htm}
  end
  str = Kramdown::Document.new(md_str, kopt).to_html
  outfile = bname + ".htm"
  File.open(outfile, "w") {|ff|  ff.print str}
  File.delete(tpl_file)  if test(?e, tpl_file)
when "word_document" then
  if defined?(Wrd)
    if (winfo = ysearch(yopt, 'run')) or  # MS-Wordを起動する
       ysearch(yopt, 'save')
      pinfo = ''
      if winfo.class == String
        pinfo = test(?f, winfo) ? file_read(winfo) : winfo
      elsif winfo.class == Array
        pinfo = winfo.join("\n") + "\n"
      end
      md_str = md_str + "\n<!-- #{pinfo} -->\n"  if pinfo != ''
      if (wsave = ysearch(yopt, 'save'))
        wsave = (wsave =~ /^\./) ? bname + wsave : wsave
      else
        wsave = bname + ".docx"
      end
      md_str = NKF.nkf("-m0s", md_str).force_encoding('Windows-31J')
      print Wrd::mkd2docx(md_str, wsave, *popt)
    else  # MS-Wordを起動しない
      docx_str = Wrd::pandoc_docx(md_str, *popt)
      File.open(bname + ".docx", "wb") {|ff|  ff.write docx_str}
    end
  else
    STDERR.puts "'wrdap' is not loaded!"
  end
when "kpdf_document" then
  if defined?(KpdfOpt)
    kpdf_o = ysearch(yopt, 'kpdf_set')
    kpdf_o = [kpdf_o]  if kpdf_o.class == String
    if kpdf_o.class == Array
      kpdf = Kramdown::kpdf
      kpdf_o.each do |kset|
        next  unless kset.class == String
        kset = kset.sub(/^\s+/, "")
        if kset =~ /^([a-z]+)[^a-z]/
          kkey = $1
          kset = (%w(dir root header codeblock codespan table) +
              %w(size dopt)).include?(kkey) ? "kpdf." + kset : kset
          eval(kset)
        end
      end
    end
    kopt.delete('template')  if kopt['template']
    str = Kramdown::Document.new(md_str).to_kpdf
    outfile = bname + ".pdf"
    File.open(outfile, "wb") {|ff|  ff.write str}
  else
    STDERR.puts "'kpdf' is not loaded!"
  end
when "pdf_document", "tex_document" then
  if (whdpt = ysearch(yopt, 'headpart'))
    hdpt = ''
    if whdpt.class == String
      hdpt = test(?f, whdpt) ? file_read(whdpt) : whdpt
    elsif whdpt.class == Array
      hdpt = whdpt.join("\n") + "\n"
    end
    $headpart += hdpt
  end
  classoption = !yopt['classoption'] ? '' : "[#{yopt['classoption']}]"
  documentclass = !yopt['documentclass'] ? '{jarticle}' :
      "{#{yopt['documentclass']}}"
  $dclass = "\\documentclass#{classoption}#{documentclass}"
  if kopt['template'] == tpl_file
    File.open(tpl_file, "w") {|ff|  ff.print tpl_tex}
  end
  str = Kramdown::Document.new(md_str, kopt).to_latex
  str = NKF.nkf("-m0w", str)
  File.delete(tpl_file)  if test(?e, tpl_file)
  ttl_str = ''
  if yopt['title']
    ttl_str += "\\title{#{conv_tex(yopt['title'])}}\n"
  end
  if yopt['author']
    ttl_str += "\\author{#{conv_tex(yopt['author'])}}\n"
  end
  if yopt['date']
    ttl_str += "\\date{#{conv_tex(yopt['date'])}}\n"
  end
  if ttl_str != ''
    if classoption !~ /slide/
      str = str.sub(/^(\\begin\{document\}.*)$/,
          "#{ttl_str}\\1\n\\maketitle")
    else  # slide指定の場合
      str = str.sub(/^(\\begin\{document\}.*)$/,
          "#{ttl_str}\\1\n\\pagenumbering{Roman}\n" +
          "\\maketitle\n\\pagenumbering{arabic}")
    end
  end
  tex_file = bname + ".tex"
  File.open(tex_file, "w") {|ff|  ff.print str}
end

if doc_type == "pdf_document"
      ## pdfの生成
  bat_file = "make_pdf.bat"
  cmd = ["@echo off"]
  cmd << "platex \"#{tex_file}\""
  cmd << "if EXIST \"#{bname}.toc\" (" +
      "platex \"#{tex_file}\")"
  cmd << "if EXIST \"#{bname}.dvi\" (" +
      "dvipdfmx \"#{bname}.dvi\")"
  File.open(bat_file, "w") {|ff|  ff.print cmd.join("\n") + "\n"}
  $stdout_old = $stdout.dup;  $stdout.reopen("!stdout.txt")
  $stderr_old = $stderr.dup;  $stderr.reopen("!stderr.txt")
  system(bat_file)
  $stdout.flush;  $stdout.reopen $stdout_old
  $stderr.flush;  $stderr.reopen $stderr_old
end

end_time = Time.now.to_f
File.open(log_file, "a") {|ff|
  ff.printf("*processing time(total): %f\n", end_time - start_time)}
output_move(bname, flist)

__END__
<script type="text/x-mathjax-config">
  MathJax.Hub.Config({
    tex2jax: {
      inlineMath: [ ['$','$'], ["\\(","\\)"] ],
      displayMath: [ ['$$','$$'], ["\\[","\\]"] ]
    }
  });
</script>
<script type="text/javascript"
  src="http://cdn.mathjax.org/mathjax/latest/MathJax.js?config=TeX-AMS_HTML">
</script>
<meta http-equiv="X-UA-Compatible" CONTENT="IE=EmulateIE7" />
####----####
<% if RUBY_VERSION >= '1.9'
     enc = @converter.root.options[:encoding]
     enc = Encoding.default_external unless enc
     cset = enc.to_s =~ /windows-31j|cp932|sjis|shift.jis/i ?
         'Shift_JIS' : enc.to_s
   else
     enc = $KCODE.to_s[0,1].downcase
     enc = 'w' unless ['e', 's'].include?(enc)
     cset = {'e'=>'euc-jp', 's'=>'Shift_JIS', 'w'=>'utf-8'}[enc]
   end
%><!DOCTYPE html>
<html>
<head>
<meta http-equiv="Content-type" content="text/html; charset=<%= cset %>" />
<% extend ::Kramdown::Utils::Html
title = ''
h = @converter.root.children.find {|c| c.type == :header}
if h
  collector = lambda {|c| c.children.collect {|cc| cc.type == :text ?
  escape_html(cc.value, :text) : collector.call(cc)}.join('')}
  title = collector.call(h)
  title = title.encode(enc) if RUBY_VERSION >= '1.9'
end %><title><%= title %></title>
<meta name="generator" content="kramdown <%= ::Kramdown::VERSION %>" />
<%= $headpart.to_s %></head>
<body>
<%= @body %>
</body>
</html>
####----####
<%= !$dclass ? "\\documentclass{jarticle}" : $dclass %>
\usepackage{listings}
\usepackage[dvipdfmx]{hyperref}
\usepackage[dvipdfmx]{color}
<% @converter.data[:packages].each do |pkg| pkg_s = ['graphicx'].include?(pkg) ? "[dvipdfmx]{#{pkg}}" : "{#{pkg}}" %>\usepackage<%= pkg_s %>
<% end %><% if @converter.data[:packages].include?('fancyvrb') %>\VerbatimFootnotes
<% end %><% if @converter.data[:packages].include?('acronym') %><%   @converter.root.options[:abbrev_defs].each_pair do |k,v| %>\acrodef{<%= @converter.normalize_abbreviation_key(k) %>}[<%= k %>]{<%= @converter.escape(v) %>}
<%   end %><% end %>\setcounter{footnote}{<%= @converter.options[:footnote_nr] - 1 %>}<%= ($headpart == '') ? '' : "\n" + $headpart %>
\hypersetup{colorlinks=true,urlcolor=blue}
\begin{document}
<%= @body %>
\end{document}
